use config::{Config, File};
use glob::glob;
use lazy_static::lazy_static;
use log::LevelFilter;
use poise::serenity_prelude as serenity;
use poise::serenity_prelude::UserId;
use regex::Regex;
use simplelog::SimpleLogger;
use std::{collections::HashSet, sync::RwLock};

// mod entity;

// Types used by all command functions
type Error = Box<dyn std::error::Error + Send + Sync>;
type Context<'a> = poise::Context<'a, Data, Error>;

mod commands;

// Custom user data passed to all command functions
pub struct Data {
    database: sqlx::SqlitePool,
}

lazy_static! {
    static ref SETTINGS: RwLock<Config> = RwLock::new(Config::default());
}

/// Register application commands in this guild or globally
///
/// Run with no arguments to register in guild, run with argument "global" to register globally.
#[poise::command(
    prefix_command,
    hide_in_help,
    // required_permissions = "MANAGE_MESSAGES | ADMINISTRATOR",
    owners_only=true,
)]
async fn register(ctx: Context<'_>, #[flag] global: bool) -> Result<(), Error> {
    // poise::builtins::register_application_commands(ctx, global).await?;

    let mut commands_builder = serenity::CreateApplicationCommands::default();
    let commands = &ctx.framework().options().commands;
    for command in commands {
        if let Some(slash_command) = command.create_as_slash_command() {
            commands_builder.add_application_command(slash_command);
        }
        if let Some(context_menu_command) = command.create_as_context_menu_command() {
            commands_builder.add_application_command(context_menu_command);
        }
    }
    let commands_builder = serenity::json::Value::Array(commands_builder.0);

    let is_bot_owner = ctx.framework().options().owners.contains(&ctx.author().id);
    if global {
        if !is_bot_owner {
            ctx.say("Can only be used by bot owner").await?;
            return Ok(());
        }

        ctx.say(format!("Registering {} commands...", commands.len()))
            .await?;
        ctx.discord()
            .http
            .create_global_application_commands(&commands_builder)
            .await?;
    } else {
        let guild = match ctx.guild() {
            Some(x) => x,
            None => {
                ctx.say("Must be called in guild").await?;
                return Ok(());
            }
        };
        let is_guild_owner = ctx.author().id == guild.owner_id;

        if !is_guild_owner && !is_bot_owner {
            ctx.say("Can only be used by server owner").await?;
            return Ok(());
        }

        ctx.say(format!("Registering {} commands...", commands.len()))
            .await?;
        ctx.discord()
            .http
            .create_guild_application_commands(guild.id.0, &commands_builder)
            .await?;
    }

    ctx.say("Done!").await?;

    Ok(())
}

async fn on_error(error: poise::FrameworkError<'_, Data, Error>) {
    // This is our custom error handler
    // They are many errors that can occur, so we only handle the ones we want to customize
    // and forward the rest to the default handler
    match error {
        poise::FrameworkError::Setup { error } => panic!("Failed to start bot: {:?}", error),
        poise::FrameworkError::Command { error, ctx } => {
            println!("Error in command `{}`: {:?}", ctx.command().name, error,);
        }
        error => {
            if let Err(e) = poise::builtins::on_error(error).await {
                println!("Error while handling error: {}", e)
            }
        }
    }
}

/// Show this help menu
#[poise::command(prefix_command, track_edits, slash_command)]
async fn help(
    ctx: Context<'_>,
    #[description = "Specific command to show help about"]
    #[autocomplete = "poise::builtins::autocomplete_command"]
    command: Option<String>,
) -> Result<(), Error> {
    poise::builtins::help(
        ctx,
        command.as_deref(),
        poise::builtins::HelpConfiguration {
            extra_text_at_bottom: "\
This bot is very much a work in progress. Commands will change and may break.",
            show_context_menu_commands: true,
            ..Default::default()
        },
    )
    .await?;
    Ok(())
}

fn setup_config() -> Result<(), Error> {
    SETTINGS
        .write()
        .unwrap()
        .merge(
            glob("data/config/*")
                .unwrap()
                .map(|path| File::from(path.unwrap()))
                .collect::<Vec<_>>(),
        )
        .unwrap();

    println!("{:#?}", SETTINGS.read());
    Ok(())
}

async fn app() -> Result<(), Error> {
    let discord = SETTINGS.read().unwrap().get_table("discord").unwrap();

    let mut owners: HashSet<UserId> = HashSet::new();
    owners.insert(UserId(117992484310745097));

    let prefix: String = discord
        .get("prefix")
        .expect("Config error, please set the discord[token] value")
        .to_owned()
        .try_into()
        .expect("Config error, please make sure discord[token] is a string");
    let options = poise::FrameworkOptions {
        commands: vec![
            help(),
            register(),
            commands::invites::invites(),
            commands::schedule::schedule(),
            commands::boop(),
            commands::fix(),
            commands::planning::get_command(),
            commands::get_links(),
        ],
        prefix_options: poise::PrefixFrameworkOptions {
            prefix: Some(prefix.into()),
            edit_tracker: Some(poise::EditTracker::for_timespan(
                std::time::Duration::from_secs(3600),
            )),
            additional_prefixes: vec![
                poise::Prefix::Regex(Regex::new("(?i)ol'? rusty please").unwrap()),
                poise::Prefix::Regex(Regex::new("(?i)Oi Rusty please").unwrap()),
            ],
            ..Default::default()
        },
        on_error: |error| Box::pin(on_error(error)),
        pre_command: |ctx| {
            Box::pin(async move {
                println!("Executing command {}...", ctx.command().qualified_name);
            })
        },
        post_command: |ctx| {
            Box::pin(async move {
                println!("Executed command {}!", ctx.command().qualified_name);
            })
        },
        owners,
        ..Default::default()
    };

    // let discord = SETTINGS.read().unwrap().get_table("discord").unwrap();
    let token = discord
        .get("token")
        .expect("Config error, please set the discord[token] value")
        .clone()
        .into_str()
        .expect("Config error, please make sure discord[token] is a string");

    let database = sqlx::sqlite::SqlitePoolOptions::new()
        .max_connections(5)
        .connect_with(
            "sqlite:data/database.db"
                .parse::<sqlx::sqlite::SqliteConnectOptions>()?
                .create_if_missing(true),
        )
        .await?;
    sqlx::migrate!("./migrations").run(&database).await?;

    poise::Framework::build()
        .token(token)
        .user_data_setup(move |_ctx, _ready, _framework| {
            Box::pin(async move { Ok(Data { database: database }) })
        })
        .options(options)
        .run()
        .await
        .unwrap();

    Ok(())
}

#[tokio::main]
async fn main() {
    print!("Starting the bot");
    setup_config().unwrap();
    // env_logger::init();
    // init_log();
    let log_default = LevelFilter::Warn;
    let logger = match SETTINGS
        .read()
        .unwrap()
        .get_table("logging")
        .unwrap_or_default()
        .get("level")
    {
        Some(val) => match val.clone().into_str() {
            Ok(level) => match level.to_lowercase().as_str() {
                "trace" => LevelFilter::Trace,
                "debug" => LevelFilter::Debug,
                "info" => LevelFilter::Info,
                "warn" => LevelFilter::Warn,
                "error" => LevelFilter::Error,
                _ => log_default,
            },
            _ => log_default,
        },
        _ => log_default,
    };

    println!("Logging level: {:?}", logger);

    SimpleLogger::init(logger, simplelog::Config::default()).unwrap();

    if let Err(e) = app().await {
        log::error!("{}", e);
        std::process::exit(1);
    }
}
